import 'dart:math';

import 'package:bitcoin_flutter/bitcoin_flutter.dart';
import 'package:web3dart/crypto.dart';

import 'walletProxy.dart';

class BitcoinIn {
  String txHash;
  int vout;
  num value;

  BitcoinIn(this.txHash, this.vout, this.value);
}

class Btransaction {
  TransactionBuilder txb;
  NetworkType network;

  Btransaction(this.txb, this.network);
}

class BitcoinTransaction {
  static Future<String> signBitcoinTransaction(
      String privateKey, Btransaction transaction) async {
    if (transaction != null) {
      try {
        ECPair ecPair =
            ECPair.fromWIF(privateKey, network: transaction.network);
        transaction.txb.inputs.forEach((element) {
          int index = transaction.txb.inputs.indexOf(element);
          transaction.txb.sign(vin: index, keyPair: ecPair);
        });
        Transaction tx = transaction.txb.buildIncomplete();
        String hex = tx.toHex();
        return hex;
      } catch (e) {
        print('BitcoinTransaction.signBitcoinTransaction error : $e');
      }
    } else {
      print(
          'BitcoinTransaction.signBitcoinTransaction error : transaction is null.');
    }
    return null;
  }

  // 6a = 106 OP_RETURN
  // 14 = 20
  // 6f6d6e69 = omni
  static const _omni = '6a146f6d6e69';
  static const int dirt = 546;
  static String _fillLengthWith(String s,
      {int length = 16, String char = '0'}) {
    if (s == null) return '';
    if (s.length >= length) return s.substring(0, length);
    return (char * length + s).substring(s.length);
  }

  static String _createOmniPayload(num value, int contract) {
    String type = contract.toRadixString(16);
    String amount = value.toInt().toRadixString(16);
    String payload = _omni + _fillLengthWith(type) + _fillLengthWith(amount);
    return payload;
  }

  static Future<Btransaction> createBitcoinTransaction(
      String from, String to, num value, num fee,
      {String type = 'BTC', List<BitcoinIn> unspends}) async {
    if (WalletProxy.bitcoinType.keys.contains(type)) {
      CoinInfo coinInfo = WalletProxy.bitcoinType[type];
      try {
        final txb = new TransactionBuilder(network: coinInfo.network);
        int coinFee = (fee * pow(10, coinInfo.decimals)).toInt();
        int coinValue = (value * pow(10, coinInfo.decimals)).toInt();
        int needAmount = coinValue + coinFee;
        if (unspends == null) {
          //just for test
          unspends = _checkUnspends(from, needAmount);
        }
        int amount = 0;
        unspends.forEach((element) {
          amount += (element.value * pow(10, coinInfo.decimals)).toInt();
          txb.addInput(element.txHash, element.vout);
        });
        int change = amount - coinValue - coinFee;
//        print(amount);
//        print(coinValue);
//        print(coinFee);
//        print(change);
        if (change < 0) {
          print(
              'BitcoinTransaction.createBitcoinTransaction error : Amount is not enough.');
        } else {
          if (change >= dirt) {
            txb.addOutput(from, change);
          }
          txb.addOutput(to, coinValue);
          return Btransaction(txb, coinInfo.network);
        }
      } catch (e) {
        print('BitcoinTransaction.createBitcoinTransaction error : $e');
      }
    }
    return null;
  }

  static Future<Btransaction> createOmniTransaction(
      String from, String to, num value, num fee,
      {String type = 'USDT', List<BitcoinIn> unspends}) async {
    if (WalletProxy.omniType.keys.contains(type)) {
      CoinInfo coinInfo = WalletProxy.omniType[type];
      try {
        final txb = new TransactionBuilder(network: coinInfo.network);
        const int decimals = 8;
        int coinFee = (fee * pow(10, decimals)).toInt();
        int needAmount = coinFee + dirt;
        if (unspends == null) {
          //just for test
          unspends = _checkUnspends(from, needAmount);
        }
        int amount = 0;
        unspends.forEach((element) {
          amount += (element.value * pow(10, decimals)).toInt();
          txb.addInput(element.txHash, element.vout);
        });
        int change = amount - dirt - coinFee;
//        print(amount);
//        print(dirt);
//        print(coinFee);
//        print(change);
        if (change < 0) {
          print(
              'BitcoinTransaction.createOmniTransaction Amount is not enough.');
        } else {
          if (change >= dirt) {
            txb.addOutput(from, change);
          }
          txb.addOutput(to, dirt);

          int coinValue = (value * pow(10, coinInfo.decimals)).toInt();
          int contract = num.parse(coinInfo.address).toInt();
          String payload = _createOmniPayload(coinValue, contract);
          print(payload);

          txb.addOutput(hexToBytes(payload), 0);

          return Btransaction(txb, coinInfo.network);
        }
      } catch (e) {
        print('BitcoinTransaction.createOmniTransaction error : $e');
      }
    }
    return null;
  }

  static List<BitcoinIn> _checkUnspends(String from, int amount) {
    List<BitcoinIn> ins = List<BitcoinIn>()
      ..add(BitcoinIn(
          '30db0e994544a63d3b6e2e990015022dc356d740612295a0e25bad533d923da3',
          1,
          0.0441695))
      ..add(BitcoinIn(
          '54ce8da7d37b34d764fb72c1626bb684aee49f8b9272989327b2b1e191d3f4aa',
          0,
          0.00000546));
    return ins;
  }
}
